#include <assert.h>
#include <bfd.h>
#include <libiberty.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include <sys/stat.h>

#include <linux/module.h>

#define LINKER_NAME		"ld"
#define LINKER_SCRIPT_NAME	"/usr/lib/ldscripts/module.lds"

#define DEBUG

#ifdef DEBUG
# define DBG(args)	fprintf args
#else
# define DBG(args)
#endif

extern int get_kernel_syms (struct kernel_sym *table);
extern int create_module (const char *module_name, unsigned long size,
			  unsigned long *addr);
extern int delete_module (char *module_name);
extern int init_module (char *module_name, char *code, unsigned codesize,
			struct mod_routines *routines, 
			struct symbol_table *symtab);

static char *prog_name, *mod_name, *file_name, *ksym_name = 0, *code;
static unsigned long mod_addr = 0, code_size;
static long num_module_refs = 0;	/* # of references to *other* modules */
static long num_global_syms = 0;	/* # of exported symbols in module */
static long string_tab_size = 0;	/* size of string table */
static int verbose = 0;


static void
usage(FILE *where)
{
    fprintf(where, "Usage: %s insmod [-f] [-x] [-o name] [-m] module\n",
	    prog_name);
}


static void
clean_exit (int exitcode)
{
    if (mod_addr) {
	delete_module (mod_name);
    }
    if (ksym_name) {
	unlink(ksym_name);
    }
    exit (exitcode);
}


static char *
tmp_name (const char *dir, const char *prefix, const char *postfix)
{
    static const char *tmp_dir = 0;
    char path[PATH_MAX];
    int pid = getpid();

    if (!tmp_dir) {
	tmp_dir = getenv("TMPDIR");
	if (!tmp_dir) {
	    tmp_dir = strdup(dir);
	    if (!tmp_dir) {
		tmp_dir = "/tmp";
	    }
	}
    }
    sprintf(path, "%s/%s%d%s", dir, prefix, pid, postfix);
    return strdup(path);
}


static int
hidden_sym (const char * name)
{
    switch (name[0]) {
      case 'c': return strcmp(name, "cleanup_module") == 0;
      case 'i': return strcmp(name, "init_module") == 0;
      case 'k': return strcmp(name, "kernel_version") == 0;
      case 'm': return strcmp(name, "mod_use_count_") == 0;
    }
    return 0;
}


static void
accum_size (bfd *abfd, asection *sect, PTR obj)
{
    unsigned long *size = (unsigned long *) obj;
    *size += (bfd_section_size (abfd, sect) +
	      (1 << bfd_get_section_alignment(abfd, sect)));
}


static void
load_module (bfd *abfd, asection *sect, PTR obj)
{
    unsigned long size = bfd_section_size (bfd, sect);
    unsigned long *total_size = (unsigned long *) obj;

    *total_size += size;
    if (sect->flags & SEC_ALLOC) {
	char * addr;

	addr = code + (bfd_get_section_vma(bfd, sect) - mod_addr);
	if (addr + size >= code + code_size) {
	    return;
	}
	if (sect->flags & SEC_LOAD) {
	    if (!bfd_get_section_contents (abfd, sect, addr, 0, size)) {
		bfd_perror ("bfd_get_section_contents");
		clean_exit (1);
	    }
	} else {
	    memset (addr, 0, size);
	}
    }
}


static int
bfd_get_symtab (bfd *abfd, asymbol ***symtabP)
{
    unsigned long storage;

    storage = bfd_get_symtab_upper_bound(abfd);
    if (!storage) {
	return -1;
    }
    *symtabP = (asymbol **) xmalloc (storage);
    if (!*symtabP) {
	return -1;
    }
    if (!bfd_canonicalize_symtab(abfd, *symtabP)) { 
	return -1;
    }
    return 0;
}


/*
 * Print a loadmap so that kernel panics can be identified.  Load map
 * option conceived by Derek Atkins <warlord@MIT.EDU>.
 */
static void
print_load_map (asymbol **symtab)
{
    asymbol **sp;

    for (sp = symtab; *sp; sp++) {
	const char *name;

	name = bfd_asymbol_name (*sp);
	if (((*sp)->flags & BSF_DEBUGGING) ||
	    bfd_is_abs_section (bfd_get_section (*sp)) ||
	    strpbrk (name, ".$") != 0)
	{
	    /*
	     * Skip debugging, kernel, and compiler-generated symbols.
	     */
	    continue;
	}
	printf("%08lx %c %s\n", (long) bfd_asymbol_value (*sp),
	       bfd_decode_symclass (*sp), name);
    }
}


static void
count_refs_and_globals (asymbol **symtab, struct kernel_sym *ksyms, long nksyms)
{
    asymbol **sp;

    for (sp = symtab; *sp != 0; ++sp) {
	if (bfd_is_und_section (bfd_get_section (*sp))) {
	    struct kernel_sym *ksym, *mod_ksym = 0;
	    int i;

	    /* it's a reference to a kernel symbol: */

	    for (ksym = ksyms, i = 0; i < nksyms; ++ksym, ++i) {
		if (ksym->name[0] == '#' || ksym->name[0] == '+') {
		    mod_ksym = ksym;
		    continue;
		}
		/* skip over `_' in ksym->name: */
		if (strcmp (bfd_asymbol_name (*sp), ksym->name + 1) == 0) {
		    if (verbose > 1) {
			DBG ((stderr, "%s: symbol `%s' defined in %s\n",
			      prog_name, bfd_asymbol_name(*sp),
			      mod_ksym->name[1] ? mod_ksym->name + 1 : "kernel"));
		    }
		    assert(mod_ksym);
		    if (mod_ksym->name[1] && mod_ksym->name[0] == '#') {
			mod_ksym->name[0] = '+';	/* mark module as referenced */
			++num_module_refs;
		    }
		    break;
		}
	    }
	} else if (((*sp)->flags & BSF_GLOBAL) &&
		   !hidden_sym(bfd_asymbol_name(*sp)))
	{
	    /* it's an exported symbol: */
	    ++num_global_syms;
	    /* leave space for leading `_' and trailing `\0': */
	    string_tab_size += strlen(bfd_asymbol_name (*sp)) + 2;
	}
    }
}


static void *
lookup_symval (asymbol **symtab, const char *name)
{
    asymbol **sp;

    for (sp = symtab; *sp != 0; ++sp) {
	if (strcmp ((*sp)->name, name) == 0) {
	    if (((*sp)->flags & BSF_GLOBAL) &&
		!((*sp)->flags & BSF_DEBUGGING))
	    {
		return (void *) bfd_asymbol_value (*sp);
	    }
	}
    }
    return 0;

}


static struct symbol_table *
create_module_symbol_table (asymbol **mod_symtab, asymbol **final_symtab,
			    struct kernel_sym *ksyms, long nksyms)
{
    struct symbol_table *symtab;
    struct module_ref *mod_ref;
    struct internal_symbol *isym;
    struct kernel_sym *ksym;
    char *string_tab;
    asymbol **sp;
    long i, size;

    size = (sizeof (struct symbol_table) +
	    num_global_syms * sizeof (struct internal_symbol) +
	    num_module_refs * sizeof (struct module_ref) +
	    string_tab_size);
    symtab = xmalloc (size);
    symtab->size = size;
    symtab->n_symbols = num_global_syms;
    symtab->n_refs = num_module_refs;

    string_tab = ((char *) symtab) + size - string_tab_size;

    isym = (struct internal_symbol *) &symtab->symbol[0];
    for (sp = mod_symtab; *sp != 0; ++sp) {
	if (!bfd_is_und_section (bfd_get_section (*sp)) &&
	    ((*sp)->flags & BSF_GLOBAL) &&
	    !hidden_sym(bfd_asymbol_name(*sp)))
	{
	    /* it's an exported symbol: */
	    isym->name = (char *) (string_tab - (char *) symtab);
	    isym->addr = lookup_symval (final_symtab, bfd_asymbol_name (*sp));
	    *string_tab++ = '_';
	    strcpy(string_tab, bfd_asymbol_name (*sp));
	    string_tab += strlen (string_tab) + 1;
	    --num_global_syms;

	    if (verbose > 1) {
		DBG ((stderr, "%s: added symbol `%s' at %p\n",
		      prog_name, ((char *) symtab) + (unsigned long) isym->name,
		      isym->addr));
	    }
	    ++isym;
	}
    }
    assert (num_global_syms == 0);
    

    /* fill in module references: */
    mod_ref = (struct module_ref *) isym;
    for (ksym = ksyms, i = 0; i < nksyms; ++ksym, ++i) {
	if (ksym->name[0] == '+') {
	    mod_ref->module = (struct module *) ksym->value;
	    mod_ref->next = 0;	/* used by kernel */
	    ++mod_ref;
	    --num_module_refs;
	}
    }
    assert(num_module_refs == 0);

    return symtab;
}


int
main (int argc, char **argv)
{
    bfd *abfd, *mod_bfd;
    asymbol **symtab, **sym, **mod_symtab;
    long nksyms, nsyms;
    unsigned long check_size;
    struct kernel_sym *ksyms, *ksym;
    struct mod_routines routines;
    char ld_command[3*PATH_MAX + 256], *module_name;
    int ch, loadmap = 0;

    prog_name = argv[0];
    while ((ch = getopt(argc, argv, "fxXo:vmh")) != EOF) {
	switch (ch)
	  {
	    case 'h': /* print usage info */
	      usage(stdout);
	      clean_exit (0);

	    case 'm': /* generate loadmap */
	      loadmap = 1;
	      break;

	    case 'o': /* set module name */
	      mod_name = optarg;
	      break;

	    case 'v': /* verbose */
	      ++verbose;
	      break;

	    case 'f': /* force loading */
	    case 'X': /* export externs */
	      fprintf(stderr, "%s: option `%c' not implemented yet\n", 
		      prog_name, ch);
	      clean_exit (1);

	    default: 
	      fprintf(stderr, "%s: unknown option '%c'\n", prog_name, ch);
	      usage(stderr);
	      clean_exit (1);
	  }
    }
    if (optind != argc - 1) {
	fprintf(stderr, "%s: bad number of arguments\n", prog_name);
	usage(stderr);
	clean_exit (1);
    }
    file_name = argv[optind];

    /* construct module name, module entry/exit routine names */
    if (!mod_name) {
	int len;
	char *p;

	p = strrchr(file_name, '/');
	if (p) {
	    p++;
	} else {
	    p = file_name;
	}
	len = strlen(p);
	if (len > 2 && strcmp(p + len - 2, ".o") == 0) {
	    len -= 2;
	} else if (len > 4 && strcmp(p + len - 4, ".mod") == 0) {
	    len -= 4;
	}
	mod_name = strdup(p);
	mod_name[len] = '\0';
    }

    if (!strpbrk(file_name, "/.")) {
	/*
	 * FILE_NAME is not absolute and has no dots in it.  Try to map
	 * it into an existing filename.
	 */
	struct stat dummy;
	const char *dir;
	char *dirlist, path[PATH_MAX];

	dirlist = getenv("MODPATH");
	if (!dirlist) {
	    dirlist = ".:/linux/modules";
	}
	dirlist = strdup(dirlist);

	dir = strtok(dirlist, ":");
	while (dir) {
	    sprintf(path, "%s/%s.o", dir, file_name);
	    if (stat(path, &dummy) == 0) {
		break;
	    }
	    sprintf(path, "%s/%s.mod", dir, file_name);
	    if (stat(path, &dummy) == 0) {
		break;
	    }
	    dir = strtok(NULL, ":");
	}
	free(dirlist);

	file_name = path;
    }
    bfd_init ();
    abfd = bfd_openr (file_name, 0);
    if (!abfd) {
	bfd_perror (file_name);
	clean_exit (1);
    }
    if (!bfd_check_format (abfd, bfd_object)) {
	bfd_perror ("bfd_check_format");
	clean_exit (1);
    }
    /* 
     * Find out how big the module is gonna be, then create the kernel
     * module (so we know the address we have to link for):
     */
    code_size = 0;
    bfd_map_over_sections (abfd, accum_size, &code_size);
    if (!bfd_close (abfd)) {
	bfd_perror ("bfd_close");
	clean_exit (1);
    }
    if (create_module (mod_name, code_size, &mod_addr) < 0) {
	perror("create_module");
	clean_exit (1);
    }
    mod_addr += sizeof(long);

    /*
     * Get current kernel symbol table (including global symbols of
     * modules already installed).
     */
    nksyms = get_kernel_syms (0);
    if (nksyms < 0) {
	if (nksyms < 0) {
	    perror ("get_kernel_syms");
	    clean_exit (1);
	}
    }
    ksyms = xmalloc (nksyms * sizeof (*ksyms));
    if (get_kernel_syms (ksyms) != nksyms) {
	perror ("get_kernel_syms inconsistency");
	clean_exit (1);
    }

    /*
     * Find out what symbols the module defines and what references it
     * makes.
     */
    mod_bfd = bfd_openr (file_name, 0);
    if (!mod_bfd) {
	bfd_perror (file_name);
	clean_exit (1);
    }
    if (!bfd_check_format (mod_bfd, bfd_object)) {
	bfd_perror (file_name);
	clean_exit (1);
    }
    if (bfd_get_symtab (mod_bfd, &mod_symtab) < 0) {
	bfd_perror (file_name);
	clean_exit (1);
    }
    count_refs_and_globals(mod_symtab, ksyms, nksyms);

    /*
     * Create an object file that contains all kernel symbols as
     * absolute symbols:
     */
    ksym_name = tmp_name("/usr/tmp", "ksym", ".o");
    abfd = bfd_openw (ksym_name, 0);
    if (!abfd) {
	bfd_perror (ksym_name);
	clean_exit (1);
    }
    bfd_set_format (abfd, bfd_object);
    bfd_set_arch_info (abfd, bfd_scan_arch (""));

    /* make one BFD symbol for each kernel symbol: */
    symtab = xmalloc ((nksyms + 2) * sizeof (*symtab));
    for (ksym = ksyms, sym = symtab; ksym < ksyms + nksyms; ++ksym) {
	if (ksym->name[0] == '#') {
	    /* name & address of a module */
	    continue;
	}
	*sym		= bfd_make_empty_symbol (abfd);
	(*sym)->name	= ksym->name + 1;	/* XXX skip `_' for now... */
	(*sym)->value	= ksym->value;
	(*sym)->flags	= BSF_GLOBAL;
	(*sym)->section	= bfd_abs_section_ptr;
	++sym;
    }
    *sym = bfd_make_empty_symbol (abfd);
    (*sym)->name = "mod_use_count_";
    (*sym)->value = mod_addr - sizeof(long);
    (*sym)->flags = BSF_GLOBAL;
    (*sym)->section = bfd_abs_section_ptr;
    nsyms = ++sym - symtab;
    *sym = 0;					/* terminate symlist */

    /* install symbol table: */
    if (!bfd_set_symtab (abfd, symtab, nsyms)) {
	bfd_perror ("bfd_set_symtab");
	clean_exit (1);
    }
    /* write the whole thing out: */
    if (!bfd_close (abfd)) {
	bfd_perror ("bfd_close");
	clean_exit (1);
    }

    /*
     * Link the module object file with kernel symbols:
     */
    module_name = tmp_name("/usr/tmp", mod_name, "");
    sprintf(ld_command, LINKER_NAME " -o %s -defsym __start=0 -N "
	    "-Ttext %#lx -T " LINKER_SCRIPT_NAME " %s %s",
	    module_name, mod_addr, file_name, ksym_name);
    if (verbose) {
	fprintf (stderr, "%s\n", ld_command);
    }
    system(ld_command);

    unlink(ksym_name);
    ksym_name = 0;

    /*
     * Now read the newly created the module into memory:
     */
    abfd = bfd_openr (module_name, 0);
    unlink(module_name);
    if (!abfd) {
	bfd_perror (module_name);
	clean_exit (1);
    }
    if (!bfd_check_format (abfd, bfd_object)) {
	bfd_perror (module_name);
	clean_exit (1);
    }
    if (bfd_get_symtab (abfd, &symtab) < 0) {
	bfd_perror (module_name);
	clean_exit (1);
    }
    routines.init    = lookup_symval (symtab, "init_module");
    routines.cleanup = lookup_symval (symtab, "cleanup_module");
    if (!routines.init) {
	fprintf (stderr, "%s: no `module_init' in %s\n", prog_name, file_name);
	clean_exit (1);
    }
    if (!routines.cleanup) {
	fprintf (stderr, "%s: no `cleanup_module' in %s\n", prog_name, file_name);
	clean_exit (1);
    }
    code = xmalloc(code_size);
    check_size = 0;
    bfd_map_over_sections (abfd, load_module, &check_size);
    if (check_size != code_size) {
	if (check_size > code_size) {
	    fprintf (stderr,
		     "%s: module is %ld bytes long (expected %ld)\n"
		     "%s: forgot to specify -dc when linking %s?\n",
		     prog_name, check_size, code_size, prog_name, file_name);
	    clean_exit(1);
	}
	if (verbose) {
	    fprintf (stderr,
		     "%s: warning: module only %ld bytes long (expected %ld)\n",
		     prog_name, check_size, code_size);
	}
    }
    if (loadmap) {
	print_load_map (symtab);
    }
    init_module(mod_name, code, code_size, &routines,
		create_module_symbol_table (mod_symtab, symtab, ksyms, nksyms));
    return 0;
}
